#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "benchmark.h"
#include "idct.h"

// Generate random coefficients in the range [-15..15].
void gen_random_coeffs(int16_t* dst, size_t len)
{
	// Standard NR LCG (we avoid rand() to get consistent behavior across platforms).
	static uint32_t seed = 1234;
	for (unsigned i = 0; i < len; ++i) {
		seed = seed * 1664525U + 1013904223U;
		if (seed >> 31) {
			dst[i] = (uint8_t)(seed >> 27) & 0x7;
		} else {
			dst[i] = -((uint8_t)(seed >> 27) & 0x7);
		}
	}
}

// Test that the input is pretty close to the reference for random inputs. 
// (If the reference funtion is given in, this becomes a simple test of its
// determinism.)
void test_random_inputs(idct_alloc_t* idct_alloc, idct_free_t* idct_free, idct_func_t* idct)
{
	int16_t coeff[DCTSIZE2]; 
	uint32_t quant[DCTSIZE2];
	uint8_t output[DCTSIZE2];
	uint8_t reference[DCTSIZE2];

	// Unit quantization (ie., no scaling).
	for (unsigned i = 0; i < DCTSIZE2; ++i) {
		quant[i] = 1;
	}
	
	void* userdata_reference = idct_reference_alloc(quant);
	void* userdata = idct_alloc(quant);

	for (unsigned i = 0; i < 1000; ++i) {	
		gen_random_coeffs(coeff, DCTSIZE2);

		(*idct)(coeff, userdata, output);
		(idct_reference)(coeff, userdata_reference, reference);

		// Find the RMS difference.
		int diff_squared = 0;
		for (unsigned i = 0; i < DCTSIZE2; ++i) {
			diff_squared += (output[i] - reference[i]) * (output[i] - reference[i]);
		}

		assert(diff_squared <= 5);
	}

	idct_reference_free(userdata_reference);
	idct_free(userdata);
}

// Test that a single DC coefficient becomes spread out to all blocks.
void test_dc_becomes_spread_out(idct_alloc_t* idct_alloc, idct_free_t* idct_free, idct_func_t* idct)
{
	int16_t coeff[DCTSIZE2] = { 0 }; 
	uint32_t quant[DCTSIZE2];
	uint8_t output[DCTSIZE2];
	
	// Unit quantization (ie., no scaling).
	for (unsigned i = 0; i < DCTSIZE2; ++i) {
		quant[i] = 1;
	}

	void* userdata = idct_alloc(quant);

	for (unsigned i = 0; i < 255*8; ++i) {	
		uint32_t reference_value = i / 8;
		coeff[0] = i;

		(*idct)(coeff, userdata, output);

		for (unsigned i = 0; i < DCTSIZE2; ++i) {
			assert(abs(output[i] - reference_value) <= 1);
		}
	}
	
	idct_free(userdata);
}

double timediff(const struct timeval* a, const struct timeval* b)
{
	return (double)(b->tv_sec - a->tv_sec) +
		(double)(b->tv_usec - a->tv_usec) * 1e-6;
}

void test_performance(idct_alloc_t* idct_alloc, idct_free_t* idct_free, idct_func_t* idct)
{
	const unsigned num_runs = (idct == idct_reference) ? 5000 : 5000000;

	int16_t coeff[DCTSIZE2]; 
	uint32_t quant[DCTSIZE2];
	uint8_t output[DCTSIZE2];
		
	gen_random_coeffs(coeff, DCTSIZE2);
	
	// Unit quantization (ie., no scaling).
	for (unsigned i = 0; i < DCTSIZE2; ++i) {
		quant[i] = 1;
	}

	void* userdata = idct_alloc(quant);

	start_benchmark_timer();

	for (unsigned i = 0; i < num_runs; ++i) {
		(*idct)(coeff, userdata, output);
	}
	
	double diff = stop_benchmark_timer();
	printf("%u runs in %.2f CPU seconds = %.2f IDCTs/sec\n",
		num_runs, diff, num_runs / diff);

	idct_free(userdata);
}

void test_all_idct(idct_alloc_t* idct_alloc, idct_free_t* idct_free, idct_func_t* idct)
{
	printf("  test_dc_becomes_spread_out()\n");
	test_dc_becomes_spread_out(idct_alloc, idct_free, idct);	

	printf("  test_random_inputs()\n");
	test_random_inputs(idct_alloc, idct_free, idct);	

	printf("  performance test: ");
	test_performance(idct_alloc, idct_free, idct);
}

int main(void)
{
	printf("idct_reference:\n");
	test_all_idct(idct_reference_alloc, idct_reference_free, idct_reference);

	printf("idct_float:\n");
	test_all_idct(idct_float_alloc, idct_float_free, idct_float);

	printf("All tests pass.\n");
	return 0;
}
